// JBBRXG11 - DX11 Effect Template
//
// This template is an outline of a shader chain: vertex, hull,
// domain, geometry and pixel shaders.  It doesn't attempt to 
// show any of the modelling that can be achieved with interesting
// tellelation patch data - the patch data expected is here simply
// three corner positions of a triangle to be subdivided.  It is
// set up for triangle tesselation.  Data should be sent using
// PrimitiveType.PatchListWith3ControlPoints.  Experimentally we
// have found that setting a shader chain including tesselation
// running with anything other than a PatchList... PrimitiveType 
// will crash the whole computer.
// Hint: Try with FillMode.WireFrame to show the tesselation.

float4x4 World;
float4x4 View;
float4x4 Projection;

// TODO: add additional effect parameters here.

//// This is the DX10/11 syntax for texture sampler declarations
//// There is no need to mention the texture here - see pixel
//// shader for sample texture sampling instruction.
//SamplerState samLinear
//{
//    Filter = MIN_MAG_MIP_LINEAR;
//    AddressU = Wrap;
//    AddressV = Wrap;
//};

struct VertexShaderInput
{
    float4 Position : POSITION;

    // TODO: add input channels such as texture
    // coordinates and vertex colors here.
};

struct VertexShaderOutput
{
    float4 Position : POSITION;

    // TODO: information to be supplied to the hull/domain shaders
};

struct HullConstantShaderOutput
{
	// Parameters to specify exterior and interior tesselation
	float edges[3]: SV_TessFactor;
	float inside: SV_InsideTessFactor;

	// TODO: add fields to provide information for Domain Shader
};

struct HullShaderOutput
{
    float4 Position : POSITION;

	// TODO: add fields to provide patch information for Domain Shader
};

struct DomainShaderOutput
{
    float4 Position : POSITION;

	// TODO: add any other fields required for geometry shader
};

struct GeometryShaderOutput
{
    float4 Position : SV_POSITION;

	// TODO: add any other fields required for pixel shader
};

VertexShaderOutput VertexShaderFunction(VertexShaderInput input)
{
    VertexShaderOutput output;
    output.Position = mul(input.Position, World);

    // TODO: add your vertex shader code here.

    return output;
}

HullConstantShaderOutput HullConstantShaderFunction(InputPatch<VertexShaderOutput, 3> inpatch, uint pid : SV_PrimitiveID)
{
	HullConstantShaderOutput output;
	output.edges[0] = output.edges[1] = output.edges[2] = 5;
	output.inside = 5;

	// TODO: Set tesselation specification(above) appropriately
	// and add any information required for Domain Shader

	return output;
}

[domain("tri")]
[partitioning("integer")]  // integer, fractional_even, fractional_odd or pow2.
[outputtopology("triangle_cw")]
[outputcontrolpoints(3)]
[patchconstantfunc("HullConstantShaderFunction")]
HullShaderOutput HullShaderFunction(InputPatch<VertexShaderOutput, 3> patch,
                                    uint item_index : SV_OutputControlPointID,
							        uint pid : SV_PrimitiveID)
{
	HullShaderOutput output;
	output.Position = patch[item_index].Position;

	// TODO: add patch information for Domain Shader

	return output;
}

[domain("tri")]
DomainShaderOutput DomainShaderFunction(HullConstantShaderOutput input,
                                        float3 location : SV_DomainLocation,  // float2 for quad
							            const OutputPatch<HullShaderOutput, 3> patch)
{
    DomainShaderOutput output;
	// Interpolate from barycentric coordinates in 'location'
    output.Position = patch[0].Position * location.x + patch[1].Position * location.y + patch[2].Position * location.z;

    // TODO:  Interpolate other channels, eg: texture coordinates

    return output;
}

[maxvertexcount(3)]
void GeometryShaderFunction(triangle DomainShaderOutput input[3], inout TriangleStream<GeometryShaderOutput> TriStream )
{
    GeometryShaderOutput output;

	for( int v=0; v<3; v++ )
	{
		output.Position = mul(input[v].Position, mul(View, Projection));
		TriStream.Append(output);
	}
	TriStream.RestartStrip();

	// TODO:  Any required geometry manipulation and addition of channels for the pixel shader
}

float4 PixelShaderFunction(GeometryShaderOutput input) : SV_TARGET
{
    // TODO: add your pixel shader code here.

    return float4(1, 0, 0, 1);

	//// Assuming a texture2D <mytexture>, as an effect parameter
	//// this is (one of) the DX10/11 way(s) of sampling
	//return <mytexture>.Sample(samLinear, <texture coordinates>);
}

technique11 Technique1
{
    pass Pass1
    {
        // TODO: set renderstates here.

        SetVertexShader( CompileShader( vs_5_0, VertexShaderFunction() ) );
		SetHullShader( CompileShader( hs_5_0, HullShaderFunction() ) );
		SetDomainShader( CompileShader( ds_5_0, DomainShaderFunction() ) );
		SetGeometryShader( CompileShader( gs_5_0, GeometryShaderFunction() ) );
        SetPixelShader( CompileShader( ps_5_0, PixelShaderFunction() ) );
    }
}
