/*  Copyright (c) 2012 William Rogers and James Boud

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using SlimDX;
using SlimDX.Direct3D11;

namespace JBBRXG11
{
    public class EffectPass
    {
        Effect parent;
        GraphicsDevice device;
        SlimDX.Direct3D11.EffectPass dxpass;

        internal EffectPass(Effect par, GraphicsDevice dev, SlimDX.Direct3D11.EffectPass dxp)
        {
            parent = par;
            device = dev;
            dxpass = dxp;
        }

        internal SlimDX.Direct3D11.EffectPass DX_EffectPass { get { return dxpass; } }
        internal Effect ParentEffect { get { return parent; } }

        public string Name
        {
            get
            {
                return dxpass.Description.Name;
            }
        }

        public void Begin()
        {
            device.CheckAndSetVertexLayout(this);
            device.SpriteBatchEffectPass = this;
            device.RenderState.ApplyUpdates();
            dxpass.Apply(device.DX_Device_Context);
        }

        public void End()
        {
            device.SpriteBatchEffectPass = null;
        }
    }
}

/*
        // Summary:
        //     Gets the set of EffectAnnotation objects for this EffectPass.
        public EffectAnnotationCollection Annotations { get; }
 */
