/*  Copyright (c) 2012 William Rogers and James Boud

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Processors;
using SlimDX;
using SlimDX.Direct3D11;
using TInput = Microsoft.Xna.Framework.Content.Pipeline.Graphics.EffectContent;
using TOutput = JBBRXG11ContentEffect.EffectInfoStuff;

namespace JBBRXG11ContentEffect
{
    [ContentProcessor(DisplayName = "JBBRXG11 Effect Processor")]
    public class JBBRXG11EffectContentProcessor : ContentProcessor<TInput, TOutput>
    {
        public override TOutput Process(TInput input, ContentProcessorContext context)
        {
            // Do a trial compilation, to check that the code is ok - will be recompiled later
            string errormessages = "";
            try
            {
                SlimDX.D3DCompiler.ShaderBytecode bc = null;
                bc = SlimDX.D3DCompiler.ShaderBytecode.Compile(input.EffectCode,
                                                               "fx_5_0",
                                                               SlimDX.D3DCompiler.ShaderFlags.None,
                                                               SlimDX.D3DCompiler.EffectFlags.None,
                                                               null,
                                                               null,
                                                               out errormessages);
                bc.Dispose();
            }
            catch (Exception e)
            {
                errormessages = e.Message;
            }

            string outmessage = "";
            int firstline = 0, firstcol = 0;
            string[] messages = errormessages.Split(new char[] { '\r', '\n' }, StringSplitOptions.RemoveEmptyEntries);
            foreach (string m in messages)
            {
                int line = 0, col = 0, at = 0, atend; bool parsed = false, iswarning = false;
                string mcopy = m.Trim();
                #region message parser
                if (at < mcopy.Length && mcopy[at] == '(')
                {
                    at++; atend = at;
                    while (atend < mcopy.Length && mcopy[atend] >= '0' && mcopy[atend] <= '9') atend++;
                    if (atend > at)
                    {
                        line = Convert.ToInt32(mcopy.Substring(at, atend - at));
                        at = atend;
                        if (at < mcopy.Length && mcopy[at] == ',')
                        {
                            at++; atend = at;
                            while (atend < mcopy.Length && mcopy[atend] >= '0' && mcopy[atend] <= '9') atend++;
                            if (atend > at)
                            {
                                col = Convert.ToInt32(mcopy.Substring(at, atend - at));
                                at = atend;
                                if (at < mcopy.Length && mcopy[at] == ')')
                                {
                                    at++;
                                    if (at < mcopy.Length && mcopy[at] == ':')
                                    {
                                        at++;
                                        while (at < mcopy.Length && mcopy[at] == ' ') at++;
                                        if (at < mcopy.Length)
                                        {
                                            parsed = true;
                                            if (mcopy[at] == 'w' || mcopy[at] == 'W') iswarning = true;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                #endregion
                if (parsed)
                {
                    if (iswarning)
                    {
                        context.Logger.LogWarning(null, new ContentIdentity("", "", line.ToString() + "," + col.ToString()),
                                                  m, new object[] { 0 });
                    }
                    else
                    {
                        if (firstline <= 0)
                        {
                            firstline = line; firstcol = col;
                        }
                        if (!outmessage.Equals("")) outmessage += "\r\n";
                        outmessage += m;
                    }
                }
                else
                {
                    if (!outmessage.Equals("")) outmessage += "\r\n";
                    outmessage += m;
                }
            }

            if (!outmessage.Equals(""))
                throw new InvalidContentException(outmessage, new ContentIdentity("", "", firstline.ToString() + "," + firstcol.ToString()), null);


            EffectInfoStuff eis = new EffectInfoStuff();
            eis.source = input.EffectCode;
            return eis;
        }
    }
}
