/*  Copyright (c) 2012 William Rogers and James Boud

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline.Processors;
using Microsoft.Xna.Framework.Content.Pipeline.Serialization.Compiler;

// TODO: replace this with the type you want to write out.
using TWrite = JBBRXG11ContentFBXModel.ModelInfoStuff;

namespace JBBRXG11ContentFBXModel
{
    [ContentTypeWriter]
    public class JBBRXG11FBXModelContentWriter : ContentTypeWriter<TWrite>
    {
        protected override void Write(ContentWriter output, TWrite value)
        {
            // ModelInfoStuff
            // Note: value.stuff can still have debugging data appended, but the
            //       first four characters must match the version string ("Vnn ")
            //value.stuff += "HelloWorld '" + value.meshes[0].parts[0].vertexelements[1].ve.Offset + "', '" + value.meshes[0].parts[0].vertexelements[1].ve.VertexElementFormat + "', '" + value.meshes[0].parts[0].vertexelements[1].ve.VertexElementUsage + "'";
            output.Write(value.stuff);
            output.Write(value.meshes.Length);
            foreach (my_mesh mesh in value.meshes)
            {
                // my_mesh
                output.Write(mesh.name);
                output.Write(mesh.sphere.Center);
                output.Write(mesh.sphere.Radius);
                output.Write(mesh.parts.Length);
                foreach (my_mesh_part part in mesh.parts)
                {
                    // my_mesh_part
                    output.Write(part.vertexOffset);
                    output.Write(part.numvertices);
                    output.Write(part.primitivecount);
                    output.Write(part.startindex);

                    output.Write(part.vertexelements.Length);
                    for (int e = 0; e < part.vertexelements.Length; e++)
                    {
                        output.Write((int)part.vertexelements[e].ve.Offset);
                        output.Write((int)part.vertexelements[e].ve.UsageIndex);
                        output.Write((int)part.vertexelements[e].ve.VertexElementFormat);
                        output.Write((int)part.vertexelements[e].ve.VertexElementUsage);
                    }

                    output.Write(part.noindices);
                    if (!part.noindices)
                    {
                        output.Write(part.computeadjacencyindices);
                        output.Write(part.use32bitindices);
                        if (!part.use32bitindices)  // 16 bit indices
                        {
                            output.Write(part.p16Indices.Length);
                            for (int i = 0; i < part.p16Indices.Length; i++)
                            {
                                output.Write(part.p16Indices[i]);
                            }
                        }
                        else                       // 32 bit indices
                        {
                            output.Write(part.p32Indices.Length);
                            for (int i = 0; i < part.p32Indices.Length; i++)
                            {
                                output.Write(part.p32Indices[i]);
                            }
                        }
                    }
                    output.Write(part.pVertices.Length);
                    output.Write(part.pVertices);
                    output.Write(part.texturekeys.Length);
                    for (int k = 0; k < part.texturekeys.Length; k++)
                        output.Write(part.texturekeys[k]);
                    output.Write(part.texturefilenames.Length);
                    for (int n = 0; n < part.texturefilenames.Length; n++)
                        output.Write(part.texturefilenames[n]);
                }
            }
            if (value.skinning == null)
            {
                output.Write(false);
            }
            else
            {
                output.Write(true);
                output.Write(value.skinning.AnimationClips.Count);
                foreach (KeyValuePair<string, AnimationClip> pair in value.skinning.AnimationClips)
                {
                    output.Write(pair.Key);
                    output.Write(pair.Value.Duration.Ticks);
                    output.Write(pair.Value.Keyframes.Length);
                    for (int k = 0; k < pair.Value.Keyframes.Length; k++)
                    {
                        // Keyframe
                        output.Write(pair.Value.Keyframes[k].Bone);
                        output.Write(pair.Value.Keyframes[k].Time.Ticks);
                        output.Write(pair.Value.Keyframes[k].Transform);
                    }
                }
                output.Write(value.skinning.BindPose.Length);
                for (int b = 0; b < value.skinning.BindPose.Length; b++)
                    output.Write(value.skinning.BindPose[b]);
                output.Write(value.skinning.InverseBindPose.Length);
                for (int b = 0; b < value.skinning.InverseBindPose.Length; b++)
                    output.Write(value.skinning.InverseBindPose[b]);
                output.Write(value.skinning.SkeletonHierarchy.Length);
                for (int s = 0; s < value.skinning.SkeletonHierarchy.Length; s++)
                    output.Write(value.skinning.SkeletonHierarchy[s]);
            }
        }

        public override string GetRuntimeReader(TargetPlatform targetPlatform)
        {
            return typeof(JBBRXG11.Model).AssemblyQualifiedName;
            //return "JBBRXG11.Model, JBBRXG11";
        }
    }
}
